/*
 *	uWeb 1.1
 *
 *	A minimal http server for small machines
 */

#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <errno.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#ifndef pdp11
#include <time.h>
#endif

#ifndef WEBROOT
#define WEBROOT "html"
#endif
#ifndef WEBPORT
#define WEBPORT 8080
#endif

static fd_set rx_set;
static fd_set tx_set;
static fd_set ex_set;

/* Client specific data for each client */
struct client {
	int fd;
	unsigned int flags;
#define CLIENT_LIVE	1
#define CLIENT_DYING	2

	/* State bytes */
	unsigned char state_1;
#define S1_LINE		1	/* Collecting an input line */
#define S1_DISCARD	2	/* Discard input */
	unsigned char state_2;
#define O1_STOP		0	/* No output to follow */
#define O1_FILE		1	/* Outputting a file */
#define O1_TEXT		2	/* Outputting text */
#define O1_TEXTFMT	3	/* Outputting a formatted text */
	unsigned char state_3;
#define HTTP_CMD	0	/* Waiting command line */
#define HTTP_PARAM	1	/* Parsing parameter lines */
#define HTTP_REPLY	2	/* Replying with header line */
#define HTTP_REPLY_TYPE	3	/* Content type */
#define HTTP_REPLY_STAT	4	/* File information */
#define HTTP_REPLY_FILE	6	/* Replying with file */

#define HTTP_ERROR_IN	100	/* Reading rest of input then sending error */
#define HTTP_ERROR	101	/* Sending an error */

	unsigned char	command;/* What HTTP are we doing */
#define CMD_NONE	0
#define CMD_GET		1
#define CMD_HEAD	2
#define CMD_POST	3
	unsigned char	file_type; /* HTTP file type */

	time_t tick;	/* Last active at */
	struct in_addr addr;

	/* Input low level state */
	char linebuf[80];	/* Used in S1_LINE to build input */
	int lbp;

	/* Output low level state */
	int out_fd;	/* Output file descriptor */
	int out_fmt;	/* Position within formatting when formatting */
	long out_len;	/* Length left to write, -1 means 'all' */
	off_t out_pos;	/* Position we are at */
	char *out_buf;	/* FIXME: union with out_pos */
	long out_param[2];/* Output parameters for formatting mode */
	
};

#define MAX_CLIENT 	16
#define MAX_IDLE	60	
#define BUSY_IDLE	5
#define TICK_GRAN	5

static char *whitespace = " \t\r\n";

/*#define ctrace(c, x)	fprintf(stderr, "++ %d:%s\n", (c->fd), (x)) */
#define ctrace(c, x)


static struct client client[MAX_CLIENT];
static int client_max_fd = -1;
static int client_run_reaper;
static int client_idle = MAX_CLIENT;
static int server_fd;

static char *file_type[] = {
	"application/octet-stream",
	"text/html",
	"image/jpeg",
	"image/gif",
	"text/plain"
};

#define FILE_TYPE_BIN	0
#define FILE_TYPE_HTML	1
#define FILE_TYPE_JPEG	2
#define FILE_TYPE_GIF	3
#define FILE_TYPE_PLAIN	4

struct ext_ent {
	char *ext;
	int type;
};

static struct ext_ent ext[] = {
	{"jpeg", FILE_TYPE_JPEG},
	{"jpg", FILE_TYPE_JPEG},
	{"gif", FILE_TYPE_GIF},
	{"html", FILE_TYPE_HTML},
	{"htm", FILE_TYPE_HTML},
	{"txt", FILE_TYPE_PLAIN},
	{"c", FILE_TYPE_PLAIN},
	{"h", FILE_TYPE_PLAIN},
	{"y", FILE_TYPE_PLAIN},
	{"lex", FILE_TYPE_PLAIN},
	{NULL, 0}
};

static int decide_file_type(name)
char *name;
{
	char *p; 
	struct ext_ent *ep = &ext[0];

	if (strcmp(name, "/") == 0)
	        return FILE_TYPE_HTML;
	
	p = strrchr(name, '.');
	if (p) {
		p++;
		while(ep->ext) {
			if (strcmp(ep->ext, p) == 0)
				return ep->type;
			ep++;
		}
		return FILE_TYPE_BIN;
	}
	return FILE_TYPE_PLAIN;
}



static void client_init()
{
	int i;
	for (i = 0; i < MAX_CLIENT; i++) {
		client[i].fd = -1;
		client[i].out_fd =-1;
	}
}

static struct client *client_alloc()
{
	int i;
	if (client_idle == 0)
		return NULL;
	for (i = 0; i < MAX_CLIENT; i++) {
		if (!(client[i].flags & CLIENT_LIVE)) {
			client[i].fd = -1;
			client[i].flags |= CLIENT_LIVE;
			client_idle--;
			return &client[i];
		}
	}
	fprintf(stderr, "idlebotch");
	return NULL;
}

static void client_set_maxfd()
{
	int sfd = server_fd;
	int i;
	for (i = 0; i < MAX_CLIENT; i++)
		if(client[i].fd > sfd)
			sfd = client[i].fd;
	client_max_fd = sfd + 1;
}


static void client_free(c)
struct client *c;
{	
	if(c->fd != -1)
		fprintf(stderr, "fdbotch");
	else
		client_idle++;
	c->fd = -1;
	c->flags &= ~CLIENT_LIVE;
	client_set_maxfd();
}

static void client_active(c)
struct client *c;
{
	c->tick = time(NULL);
}


/*
 *	Establish our listening socket
 */
static int tcp_listener(port)
int port;
{
	struct sockaddr_in self;
	int fd;
	int one = 1;

	fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (fd < 0) {
		perror("socket");
		exit(1);
	}

	self.sin_port = htons(port);
	self.sin_family = AF_INET;
	self.sin_addr.s_addr = INADDR_ANY;

	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)))
		perror("setsockopt");

	if (ioctl(fd, FIONBIO, (char *)&one) == -1) {
		perror("ioctl");
		exit(1);
	}
	if (bind(fd, (struct sockaddr *)&self, sizeof(self)) == -1) {
		perror("bind");
		exit(1);
	}
	if (listen(fd, 5) == -1) {
		perror("listen");
		exit(1);
	}
	FD_SET(fd, &rx_set);
	FD_SET(fd, &ex_set);
	return fd;
	
}

static int writestr(fd, p)
int fd;
char *p;
{
	return write(fd, p, strlen(p));
}

/*
 *	Invoked for each new client. 
 */
static struct client *client_new(fd)
int fd;
{
	struct sockaddr_in remote;
	struct client *c;
	int size = sizeof(remote);
	int one = 1;
	int rfd = accept(fd, (struct sockaddr *)&remote, &size);
	if (rfd == -1)
		return NULL;
	if (ioctl(rfd, FIONBIO, (char *)&one) == -1) {
		perror("ioctl");
		close(rfd);
		return NULL;
	}
	c = client_alloc();
	if (c == NULL) {
		close(rfd);
		return NULL;
	}
	FD_SET(rfd, &rx_set);
	FD_SET(rfd, &ex_set);
	FD_SET(rfd, &tx_set);	/* Testing */
	c->fd = rfd;
	c->lbp = 0;
	c->state_1 = S1_LINE;
	c->state_2 = O1_STOP;
	c->command = CMD_NONE;
	c->state_3 = HTTP_CMD;
	client_active(c);
	ctrace(c, "new client");
	client_set_maxfd();
	return c;
}

static void client_destroy_action(c)
struct client *c;
{
	FD_CLR(c->fd, &rx_set);
	FD_CLR(c->fd, &tx_set);
	FD_CLR(c->fd, &ex_set);
	close(c->fd);
	if (c->out_fd != -1) {
		close(c->out_fd);
		c->out_fd = -1;
	}
	ctrace(c,"destroy");
	c->fd = -1;
	c->flags &= ~CLIENT_DYING;
	client_free(c);
}

static void client_destroy(c)
struct client *c;
{
	c->flags |= CLIENT_DYING;
	client_run_reaper = 1;
}

static void client_reaper()
{
	int i;
	if (client_run_reaper == 0)
		return;
	for (i = 0; i < MAX_CLIENT; i++) {
		struct client *c = &client[i];
		if (c->flags & CLIENT_DYING)
			client_destroy_action(c);
	}
	client_run_reaper = 0;
}

/* Called when the client discovers it cannot write data it
   wishes too */
static void client_waitw(c)
struct client *c;
{
	FD_SET(c->fd, &tx_set);
	ctrace(c, "waitw");
}

/* Called when the client write data is flushed */
static void client_wdone(c)
struct client *c;
{
	FD_CLR(c->fd, &tx_set);
	ctrace(c, "wdone");
}

/* Called when the client input is to be ignored */
/* FIXME: this acts on future input but not existing received data. Fixing
   that problem is *hard* because we've got read but unconsumed data in
   the pipeline and nowhere to shove it. Making the input never read
   more bytes than can be rammed into the input buffer by some means and
   changing how its processed would do the trick. Even then you'd need
   to add extra wakeup magic to the main loop in case all the read data
   is now in the buffer and there is no file based wakeup.

   For HTTP without persistence we don't need it so simply be aware we are
   ignoring the issue very carefully */
static void client_throttle(c)
struct client *c;
{
	FD_CLR(c->fd, &rx_set);
}

/* Called to continue accepting input from this client */
static void client_unthrottle(c)
struct client *c;
{
	FD_SET(c->fd, &rx_set);
}

/* Begin the s1 state machine for sending a file to the client */
static void client_sendfile(c, fd)
struct client *c;
int fd;
{
	c->out_fd = fd;
	c->out_pos = 0;
	c->out_len = 0x7FFFFFFFL;
	c->state_2 = O1_FILE;
	client_waitw(c);
	ctrace(c, "sendfile");
}

/* Begin the s1 state machine for sending text to the client. The
   text object must have a lifespan exceeding the completion event.

   In practice we keep them static */
static void client_sendtext(c, buf)
struct client *c;
char *buf;
{
	c->out_buf = buf;
	c->out_len = strlen(buf);
	c->state_2 = O1_TEXT;
	client_waitw(c);
	ctrace(c, "sendtext");
}

/* Similar to sendtext. The l1 and l2 values can be dynamic, they are
   two copied longs. The text for dates etc is recreated and chopped if
   it turns out to be needed twice across a split write */
static void client_sendtextfmt(c, buf, l, l2)
struct client *c;
char *buf;
long l;
long l2;
{
	c->out_buf = buf;
	c->out_len = strlen(buf);
	c->out_param[0] = l;
	c->out_param[1] = l2;
	c->state_2 = O1_TEXTFMT;
	client_waitw(c);
	ctrace(c, "sendtextfmt");
}

static void client_reply_file();
static void client_reply_stat();
static void client_reply_type();

/* We have finished the write operation. Kick off the next one according to
   the state_3 machine */
static void client_write_done(c)
struct client *c;
{

	switch(c->state_3) {
		case HTTP_REPLY:
			client_reply_type(c);
			break;
		case HTTP_REPLY_TYPE:
			client_reply_stat(c);
			break;
		case HTTP_REPLY_STAT:
			client_reply_file(c);
			break;
		default:
			/* We don't support keepalive */
			client_unthrottle(c);
/*			client_shutdown(c); */
			client_destroy(c);
			break;
	}
}

static int client_write_helper(c, buf, g)
struct client *c;
char *buf;
int g;
{
	int w;

	w = write(c->fd, buf, g);
	if (w <= 0) {
		if (w == -1 && (errno == ENOBUFS || errno == EWOULDBLOCK))
			return 0;
		client_destroy(c);
		return -1;
	}

	c->out_pos += w;
	c->out_len -= w;
	c->out_buf += w;	/* No need on file but harmless */
	return w;
}

/* Return a suitable string for the formatting. This isn't stored so can
   be static. We use the file buffer passed in because it's there */
static char *client_format(c, buf)
struct client *c;
char *buf;
{
	struct tm *tm;
	long v = c->out_param[0];
	if (c->out_buf[1] == '2')
		v = c->out_param[1];

	switch(c->out_buf[2]) {
	case 'd':
		sprintf(buf, "%ld", v);
		return buf;
	case 't':
		tm = gmtime(&v);
		strftime(buf, 512, "%a, %d %b %Y %H:%M:%S GMT", tm);
		return buf;
	default:
		return "BAD FORMAT";
	}
}

	
/* Handle a write event for the client - this is an invite
   to push data */
static void client_write_event(c)
struct client *c;
{
	int l,g,w;
	/* This must be at least about 512 bytes, if you use less then the
	   2.11 BSD TCP stack reports an error of -1, ENOBUFS but actually
	   commits some of the data) */
	char buf[512];
	char *tp;
	ctrace(c, "write_event");
	client_active(c);
	switch(c->state_2) {
		case O1_STOP:
			client_wdone(c);
			break;
		case O1_FILE:
			l = sizeof(buf);
			if (l > c->out_len)
				l = c->out_len;
			g = read(c->out_fd, buf, l);
			if (g) {
				w = client_write_helper(c, buf, g);
				if (w == -1)
					return;
				if (w < g)
					if (lseek(c->out_fd, c->out_pos, L_SET) < 0)
                                        	perror("lseek");
			} else {
				close(c->out_fd);
				c->out_fd = -1;
				c->state_2 = O1_STOP;
				client_wdone(c);
				client_write_done(c);
			}
			break;
		case O1_TEXTFMT:
			/* We do some gymnastics here to avoid having
			   a formatting buffer for each client. There
			   isn't a good reason for this except that it's
			   fun */
			g = c->out_len;
			tp = strchr(c->out_buf, '%');
			/* We are finishing the buffer off */
			if (tp == NULL)
				goto twrite;
			/* Write up to the '%' */
			g = tp-c->out_buf;
			if (g)
				goto twrite;
			/* We are on the format mark */
			tp = client_format(c, buf);
			/* We may have sent bits already c->out_fmt */
			g = strlen(tp) - c->out_fmt;
			/* Write the rest of the formatted block */
			w = write(c->fd, tp + c->out_fmt, g);
			if (w <= 0) {
				client_destroy(c);
				return;
			}
			c->out_fmt += w;
			/* Incomplet - wait for a wakeup */
			if (c->out_fmt != g)
				break;
			c->out_buf += 3;	/* Skip %nx */
			c->out_len -= 3;
			/* We can try and stuff the rest in */
			break;
		case O1_TEXT:
			g = c->out_len;
twrite:
			/* We are not mid parameter - not really relevant
			   here but it's the handy place to set it back */
			c->out_fmt = 0;	
			w = client_write_helper(c, c->out_buf, g);
			if (w == -1) {
				return;
			}
			if (c->out_len == 0) {
				client_wdone(c);
				c->state_2 = O1_STOP;
				client_write_done(c);
			}
			break;
		default:
			fprintf(stderr,"s2botch");
			exit(1);
	}
}

static int open_error(code)
char *code;
{
	char buf[16];
	sprintf(buf, "../err/%s", code);
	return open(buf, O_RDONLY);
}

static int dequote(c, l)
unsigned char *c;
char l;
{
	
	if (l >= '0' && l <= '9') {
		*c = l - '0';
		return 0;
	}
	if (l >= 'a' && l <= 'f') {
		*c = l - 'a' + 10;
		return 0;
	}
	if (l >= 'A' && l <= 'F') {
		*c = l - 'A' + 10;
		return 0;
	}
	return -1;
}

static int open_with_care(path)
char *path;
{
	char *ep = path;
	char *wp = path;
	unsigned char c;
	int fd;
	struct stat st;

	/* De-quote */
	while(*ep) {
		c = *ep++;
		if (c == '%') {
			/* Quoting */
			c = 0;
			if (dequote(&c, *ep++))
				return -1;
			c <<= 4;
			if (dequote(&c, *ep++))
				return -1;
		}
		*wp++ = c;	/* May be a no-op which is fine */
	}
	*wp = 0;
	/* Check for ../ path escapes */
	if (strstr(path,"../"))
		return -1;
	/* URI should start with a / */
	if (*path != '/')
		return -1;	/* Invalid */
	/* Sop to compatibility for the top dir only */
	if (strcmp(path, "/") == 0)
		fd = open("index.html", O_RDONLY);
	else
	/* Open the rest of it relative to our cwd */
		fd = open(path + 1, O_RDONLY);
	if (fd == -1)
		return -1;
	/* Check it is regular */
	if (fstat(fd, &st) == 0)
		if ((st.st_mode & S_IFMT) == S_IFREG)
			return fd;
	close(fd);
	return -1;
}

static void http_bad(c, code)
struct client *c;
char *code;
{
	client_throttle();
	c->out_fd = open_error(code);
	if (c->out_fd == -1) {
		fprintf(stderr, "Error file %s missing.\n", code);
		client_destroy(c);
		return;
	}
	c->state_3 = HTTP_ERROR_IN;
}

/* Begin the reply, we will work through a few states sending the lines out */
static void client_reply(c)
struct client *c;
{
	time_t t = time(NULL);
	/* Work through the reply states */
	c->state_3 = HTTP_REPLY;
	client_sendtextfmt(c, "HTTP/1.1 200 OK\r\nConnection: close\r\nDate: %1t\r\nContent-type: ",(long)t, 0L);
}

static void client_reply_type(c)
struct client *c;
{
	c->state_3 = HTTP_REPLY_TYPE;
	client_sendtext(c, file_type[c->file_type]);
}

/* Now send the file information */
static void client_reply_stat(c)
struct client *c;
{
	struct stat st;
	off_t size;
	c->state_3 = HTTP_REPLY_STAT;
	if (fstat(c->out_fd, &st) < 0)
		perror("fstat");
	if (st.st_size <= 0L)
		client_sendtextfmt(c, "\r\nLast Modified: %1t\r\n\r\n", (long)st.st_mtime);
	else
		client_sendtextfmt(c, "\r\nLast-Modified: %1t\r\nContent-Length: %2d\r\n\r\n", (long)st.st_mtime, (long)st.st_size);
}


/* The reply text is sent, send the file - or finish in HEAD mode */
static void client_reply_file(c)
struct client *c;
{
	if (c->command == CMD_HEAD) {
		client_destroy(c);
		return;
	}
	c->state_3 = HTTP_REPLY_FILE;
	client_sendfile(c, c->out_fd);
}

/* A line of ascii has been accumulated and the client is throttled */
static void client_s1_line(c)
struct client *c;
{
	int fd;
	char *np;
	char *cmd;
	char *path;
	char *ver;

	ctrace(c, "s1_line");
	c->lbp = 0;
	
	np = strchr(c->linebuf, '\r');
	if (np)
		*np = 0;

	/* We have a line of something. Decide what we do next */
	switch (c->state_3) {
	case HTTP_CMD:
		cmd = strtok(c->linebuf, whitespace);
		if (cmd == NULL)
			goto bad400;
		path = strtok(NULL, whitespace);
		if (path == NULL)
			goto bad400;
		ver = strtok(NULL, whitespace);
		if (ver == NULL)
			goto bad400;
		/* FIXME: should check for 0.9 and skip waiting stuff */
		if (strncmp(ver, "HTTP/",5))
			goto bad400;
		if (strcmp(cmd, "GET") == 0)
			c->command = CMD_GET;
		else if (strcmp(cmd, "HEAD") == 0)
			c->command = CMD_HEAD;
		else {
			http_bad(c, "501");
			return;
		}
		/* So far so good */
		/* FIXME: should 415 an overlong entry here but in truth
		   we'll lose the HTTP/ and throw up anyway */
		/* Now open the file, so we can throw away the
		   pathname */
		c->out_fd = open_with_care(path, O_RDONLY);
		if (c->out_fd == -1) {
			http_bad(c, "404");
			return;
		}
		c->file_type = decide_file_type(path);
		/* Ready to parse the parameters */
		c->state_3 = HTTP_PARAM;
		ctrace(c, "move to param");
		client_unthrottle(c);
		return;
	case HTTP_PARAM:
		if (*c->linebuf == 0) {
			ctrace(c, "move to reply");
			/* Blank line - commence reply */
			c->state_3 = HTTP_REPLY;
			client_reply(c);
			/* Keep throttled */
			return;
		}
		/* Cookies, accept and other crap */
		/* Fingers in ears, sing loudly */
		client_unthrottle(c);
		return;
	case HTTP_ERROR_IN:
		if (*c->linebuf == 0) {
			ctrace(c, "move to reply");
			/* Blank line - commence reply */
			c->state_3 = HTTP_ERROR;
			/* Keep throttled */
			client_sendfile(c, c->out_fd);
			return;
		}
		/* Cookies, accept and other crap */
		/* Fingers in ears, sing loudly */
		client_unthrottle(c);
		return;
	case HTTP_REPLY:
	case HTTP_ERROR:
		/* Throttled in this state so a reply is not wanted but
		   could arrive before we throttle. If so we can happily
		   discard it as we don't do keepalive */
		return;
	default:
		fprintf(stderr, "s3botch %d\n", c->state_3);
		exit(1);
	}
bad400:
	http_bad(c, "400");
}

static void client_read_event(c)
struct client *c;
{
	char buf[80];
	int l;
	int p;

	ctrace(c, "read_event");

	client_active(c);

	l = read(c->fd, buf, 80);
	
	if (l == 0 || (l < 0 && errno != EWOULDBLOCK)) {
		client_destroy(c);
		return;
	}

	p = 0;
	while(p < l) {
		switch(c->state_1) {
		case S1_LINE:
			if (buf[p] == '\n') {
				c->linebuf[c->lbp] = 0;
				client_throttle(c);
				client_s1_line(c);
			}
			else if (c->lbp < sizeof(c->linebuf))
				c->linebuf[c->lbp++] = buf[p];
			p++;
			break;
		case S1_DISCARD:
			p = l;
			break;
		default:
			fprintf(stderr, "s1 botch\n");
			exit(1);
		}
	}
}

static void client_except_event(c)
struct client *c;
{
	ctrace(c, "exception");
	client_destroy(c);
}

static void client_tick(now)
time_t now;
{
	struct client *c;
	int i;
	for (i = 0; i < MAX_CLIENT; i++) {
		c = &client[i];
		if (c->fd == -1)
			continue;
		if ((now - c->tick) > MAX_IDLE)
			client_destroy(c);
		if (client_idle == 0 && (now - c->tick) > BUSY_IDLE)
			client_destroy(c);
	}
}

int main(argc, argv)
int argc;
char *argv[];
{
	fd_set rv,wv,ev;
	struct timeval tv;
	int i;
	int fd;
	struct client *c;
	time_t last_tick = time(NULL);
	time_t this_tick;
	
	if (chdir(WEBROOT)) {
	        perror("html");
                exit(1);
        }

	signal(SIGPIPE, SIG_IGN);

	server_fd = tcp_listener(WEBPORT);

	FD_ZERO(&rx_set);
	FD_ZERO(&tx_set);
	FD_ZERO(&ex_set);

	client_init();
	client_set_maxfd(server_fd);

	while(1) {
		rv = rx_set;
		wv = tx_set;
		ev = ex_set;

		tv.tv_sec = 5;
		tv.tv_usec = 0;

		/* Don't bother with taking new connections if
		   we are busy */
		if (client_idle) {
			FD_SET(server_fd, &rv);
			FD_SET(server_fd, &ev);
		}

		if (select(client_max_fd, &rv, &wv, &ev, &tv) < 0) {
			perror("select");
			exit(1);
		}
		if (FD_ISSET(server_fd, &rv))
			client_new(server_fd);

		for(i = 0; i < MAX_CLIENT; i++) {
			c = &client[i];
			fd = c->fd;
			if (fd == -1)
				continue;
			if (FD_ISSET(fd, &rv))
				client_read_event(c);
			if (FD_ISSET(fd, &wv))
				client_write_event(c);
			if (FD_ISSET(fd, &ev))
				client_except_event(c);
		}
		this_tick = time(NULL);
		if (this_tick > last_tick + TICK_GRAN) {
			last_tick = this_tick;
			client_tick(this_tick);
		}
		/* Throw out anything that finished this cycle */
		client_reaper();
	}
}
